"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.toHex = exports.pushConstant = exports.constantByteSize = exports.tokenize = exports.reduceOpSpaces = exports.reduceSpaces = void 0;
const opcodes_1 = require("./opcodes");
const simple_pure_utils_1 = require("simple-pure-utils");
function removeComments(code) {
    const multiLineComment = /\/\*.*\*\//g;
    const singleLineComment = /\/\/.*/g;
    return code.replace(multiLineComment, "").replace(singleLineComment, "");
}
function reduceSpaces(code) {
    const space = /(\s|\n)+/g;
    return code.replace(space, " ");
}
exports.reduceSpaces = reduceSpaces;
function reduceOpSpaces(code) {
    const r1 = /[^\S\r\n]*([(,:{])[^\S\r\n]*/g;
    code = code.replace(r1, "$1");
    const r2 = /[^\S\r\n]*(\)})/g;
    code = code.replace(r2, "$1");
    return code;
}
exports.reduceOpSpaces = reduceOpSpaces;
function tokenize(code) {
    return (0, simple_pure_utils_1.flatten)(reduceSpaces(reduceOpSpaces(removeComments(code)))
        .split(" ")
        .filter(x => x != "")
        .map(tokenizePart));
}
exports.tokenize = tokenize;
function constantByteSize(hex) {
    return Math.ceil((hex.length - 2) / 2);
}
exports.constantByteSize = constantByteSize;
function tokenizeOpCode(code) {
    const hex = opcodes_1.opCodes[code.toLowerCase()];
    if (hex != undefined) {
        return {
            type: "op",
            name: code.toLowerCase(),
            hex: hex,
            size: constantByteSize(hex)
        };
    }
    return null;
}
function tokenizeLabelDef(code) {
    {
        //single line
        const regex = /^([a-z][a-z\d_]*):$/i;
        const match = regex.exec(code);
        if (match != null) {
            return [
                {
                    type: "labelBegin",
                    name: match[1]
                },
                {
                    type: "labelEnd",
                },
                throwIfNull(tokenizeOpCode("jumpdest")),
            ];
        }
    }
    {
        // multi line
        const regex = /^([a-z][a-z\d_]*):\s*assembly\s*{$/i;
        const match = regex.exec(code);
        if (match != null) {
            return [
                {
                    type: "labelBegin",
                    name: match[1]
                }
            ];
        }
    }
    if (code == "}") {
        return [
            {
                type: "labelEnd"
            }
        ];
    }
    ;
    return null;
}
function throwIfNull(x) {
    if (x == null)
        throw new Error("null arg");
    return x;
}
function tokenizeCall(code) {
    const callRegex = /^([^\(\)]+)\((.*)\)$/;
    const callMatch = callRegex.exec(code);
    if (callMatch != null) {
        const func = callMatch[1];
        const params = callMatch[2];
        const funcTokens = tokenize(func);
        const pTokens = tokenizeParams(params);
        return [
            ...pTokens,
            ...funcTokens
        ];
    }
    return null;
}
function tokenizeLabelOp(code) {
    const regex = /^(dataSize|dataOffset)\((.+)\)$/i;
    const match = regex.exec(code);
    if (match != null) {
        const op = match[1];
        const label = match[2];
        return {
            type: "labelOp",
            label,
            op: op
        };
    }
    return null;
}
function pushConstant(hex) {
    const size = constantByteSize(hex);
    return [
        throwIfNull(tokenizeOpCode("push" + size)),
        {
            type: "const",
            hex: hex.toLowerCase(),
            size
        }
    ];
}
exports.pushConstant = pushConstant;
function tokenizeAuxData(code) {
    const prefix = "auxdata:";
    if (code.startsWith(prefix)) {
        const hex = code.substr(prefix.length);
        return [
            throwIfNull(tokenizeOpCode("invalid")),
            {
                type: "const",
                hex: hex,
                size: constantByteSize(hex)
            }
        ];
    }
    return null;
}
function toHex(num, size) {
    const h = num.toString(16);
    const pad = (size * 2) - h.length;
    return "0x" + "0".repeat(pad) + h;
}
exports.toHex = toHex;
function tokenizePart(part) {
    const opCode = tokenizeOpCode(part);
    if (opCode != null) {
        return [opCode];
    }
    const hexRegex = /^0x[\da-f]+$/i;
    if (hexRegex.test(part)) {
        return pushConstant(part);
    }
    const auxDataToken = tokenizeAuxData(part);
    if (auxDataToken != null) {
        return auxDataToken;
    }
    const labelOpToken = tokenizeLabelOp(part);
    if (labelOpToken != null) {
        return [labelOpToken];
    }
    const callToken = tokenizeCall(part);
    if (callToken != null) {
        return callToken;
    }
    const labelDefToken = tokenizeLabelDef(part);
    if (labelDefToken != null) {
        return labelDefToken;
    }
    return [{
            type: "labelOp",
            op: "dataOffset",
            label: part
        }];
}
function tokenizeParams(params) {
    return (0, simple_pure_utils_1.flatten)(params.split(/,\s*/).reverse().map(tokenize));
}
